/***************************************************************************
 *   Copyright (C) 2004 by Michael Moritz                                  *
 *   mimo@restoel.net                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "db.h"
#include "defs.h"
//#include <signal.h>
#include <syslog.h>
#include <iostream>
#include <iomanip>
#include "triplet.h"
#include <sstream>
#include <stdexcept>
#include "cfg.h"
#include "tmplwlmod.h"
#include "wldb.h"
#include "wlcacheddb.h"
#include "dbdefs.h"
#include "dbiquote.h"

#ifndef OLDWEAK
#warning "using new db scheme -- make sure you convert your db before installing this version of gps"
#endif


using namespace std;
// define the wl modules
typedef TmplWlMod<WLDB> DbWlModule;
typedef TmplWlMod<WlCachedDB> CachedDbWlModule;

DB::DB() throw(std::exception)
	: _con(0)
{
	if(dbi_initialize(0) <= 0)
		throw runtime_error("DB: no dbi drivers installed, or none found");
	bool verbose = g_getCfg().isVerbose();
	const char* dbdriver = g_getCfg().getDbType().c_str();
	bool initmode = g_getCfg().isModeInit();
	_con = dbi_conn_new(dbdriver);
	if(!_con) {
		stringstream stm;
		stm << "DB: connection/driver " << dbdriver << " could not be opened\nAvailable drivers:\n";
		list_drivers(stm);
		throw runtime_error(stm.str());
	} else {
		if(verbose)
			syslog(LOG_DEBUG,"connecting to DB, using driver %s",dbdriver);
	}
	const ParamVector& params = g_getCfg().getDbParamVector();
	for(ParamVector::const_iterator itr=params.begin(); itr != params.end();++itr) {
		// note: dbi_conn_set_option 2nd arg is non-const
		const char *key = (*itr).c_str();
		++itr;
		if(itr != params.end()) {
			if((*itr).length() < (tempCharBuf-1) ) {
				const char *value = (*itr).c_str();
				char szBuf[tempCharBuf];
				strcpy(szBuf,value);
				if(verbose) {
					if(string(key) == "password") 
						syslog(LOG_DEBUG,"setting DB option: %s to: (hidden)",key);
					else
						syslog(LOG_DEBUG,"setting DB option: %s to: %s",key,szBuf);
						
				}
				if(dbi_conn_set_option(_con, key, szBuf) != 0)
					throw runtime_error( getConError("dbi_conn_set_option "+(*itr)) );
			}
		}
	}
	if(dbi_conn_connect(_con) != 0)
		throw runtime_error( getConError("dbi_conn_connect") );
	if(verbose)
		syslog(LOG_DEBUG,"connected to DB");	
	try {
		// initialise whitelisting modules	
		const Cfg::WLMap& wlMap = g_getCfg().getWLMap();
		for(Cfg::WLMap::const_iterator itr=wlMap.begin(); itr != wlMap.end(); ++itr) {
			bool created = false;
			const std::string& tableName = itr->first;
			//cout << "tableName: " << tableName << endl;
			const WLStaticDef& wldef = g_getWlDef(tableName); //throws if not found
			switch(itr->second.getMode()) {
				case Cfg::WLEntry::WL_db:
					_wlMods.push_back(new DbWlModule (*this, initmode, wldef) );
					created = true;
					break;
				case Cfg::WLEntry::WL_dbcached:
					_wlMods.push_back(new CachedDbWlModule (*this, initmode, wldef) );
					created = true;
					break;
				case Cfg::WLEntry::WL_off:
					break;
				default:
					throw runtime_error("wl mode not supported for table: "+tableName+". Check your configuration file.");
			}
			if(verbose && created)
				syslog(LOG_DEBUG,"created wl module: %s, type: %s",
			 		tableName.c_str(),g_getCfg().getWlModeName(itr->second.getMode()).c_str());	
		}
	} catch(runtime_error &e) { //catch exception, dont exit, wl failure is nonfatal, but log error
		syslog(LOG_ERR,"Error while initialising wl module: %s",e.what());
	}
	// initialise triplets table
	if(initmode)
		createTable(g_getTripletDef());
}
void DB::close()
{
//	s_db_object = 0;
	for(WlModuleList::const_iterator itr=_wlMods.begin(); itr != _wlMods.end();++itr) 
		delete (*itr);	
	_wlMods.clear();	
	if(_con) {
//        	signal (SIGTERM, SIG_DFL);
		syslog(LOG_INFO,"disconnecting from DB");	
		dbi_conn_close(_con);
		_con = 0;
		dbi_shutdown();
	}
}


std::string DB::getConError(const std::string& pre) const
{
	if(_con != 0) {
		const char *errBuff;
		dbi_conn_error(_con,&errBuff);
		return "DB: "+pre+": "+string(errBuff);
	} else {
		return "DB: "+pre+": no connection to db";
	}
}
void DB::list_drivers(ostream &os)
{
	dbi_driver drv = 0;
	while((drv=dbi_driver_list(drv)) != 0) {
		os << " " << dbi_driver_get_name(drv);
	}
}
const std::string DB::trimLastByte(std::string str)
{
	string::size_type lastdotpos = str.rfind('.');
	if(lastdotpos != string::npos) {
		str.erase(lastdotpos,string::npos);
		str += '.';
	}
	return str;
}
//TODO optimise this sql query 
//		nogo: REPLACE because =DELETE+INSERT
//		maybe: http://dev.mysql.com/doc/mysql/en/HANDLER.html
#ifndef OLDWEAK
DB::RecordStatus DB::update(const Triplet& triplet) throw(std::exception)
{
// unlikely to happen
//	if(!_con)
//		throw runtime_error(getConError("update"));
	dbi_driver dbidrv = dbi_conn_get_driver(_con);
	DBIQuote sender(dbidrv, triplet.getSender().substr(0,MAX_LEN_SENDER));
	DBIQuote recipient(dbidrv, triplet.getRecipient().substr(0,MAX_LEN_RECIPIENT));
	DBIQuote client(dbidrv, triplet.getClientAddress().substr(0,MAX_LEN_CLIENT));
	string clientNumeric = triplet.getClientAddressNumeric();
//	bool wl_network = g_getCfg().isWlDb(TABLE_NETWORK);
	bool weak = g_getCfg().isModeWeak();
	bool verbose = g_getCfg().isVerbose();
	RecordStatus recStatus = unknown;
	time_t ltime;
	dbi_result result;
	
	if(clientNumeric.size() != IPADDR_NUM)
		throw runtime_error("Invalid IP Address: "+triplet.getClientAddress()
			+ " parsed: "+triplet.getClientAddressNumericStr());
//	DBIQuote weakclient(dbidrv,((weak||wl_network)?trimLastByte(triplet.getClientAddress().substr(0,MAX_LEN_CLIENT)):""));
//	DBIQuote weakclient(dbidrv,trimLastByte(triplet.getClientAddress().substr(0,MAX_LEN_CLIENT)));
	// check triplet against wl modules
	if(!_wlMods.empty()) {
		Triplet quotedTriplet(client(),sender(),recipient(),clientNumeric);
		for(WlModuleList::const_iterator itr=_wlMods.begin(); itr != _wlMods.end();++itr) 
			if( (*itr)->check(quotedTriplet) )
				return registered;
	}
	//
	unsigned clientIpBytes = clientNumeric.size();
	stringstream sqlSelect;
	stringstream sqlWhere;
	//construct the sql select string
	sqlSelect << "SELECT count,uts FROM "TABLE_TRIPLETS" WHERE ( ";
	sqlWhere << ATTR_SENDER"=" << sender() << " AND "ATTR_RECIPIENT"=" << recipient();
//	string sqlOrder;
	if(weak) { // weak greylisting
		clientIpBytes = g_getCfg().getWeakBytes();
		if( (clientIpBytes < 1) || (clientIpBytes > clientNumeric.size()) )
			throw runtime_error("Invalid client IP bytes (weakbytes in config): "
				+numToStr(clientIpBytes)+" (1-"+numToStr(clientNumeric.size())+"): "+triplet.getClientAddress()
				+ " parsed: "+triplet.getClientAddressNumericStr());
	}
	sqlWhere << " AND "	<< triplet.getClientAddressNumericStrSQL(" AND ","ip",64,"=",2,clientIpBytes);	
	sqlSelect << sqlWhere.str() << " )"; ///+ sqlOrder + " LIMIT 1"
	if(verbose)
			syslog(LOG_DEBUG,sqlSelect.str().c_str());

	result = DBI_CON_QUERYF(_con, sqlSelect.str().c_str());
	if(!result)
		throw runtime_error(getConError("update select: ")+sqlSelect.str() );

	time(&ltime); //get ultime (unix timestamp) in ltime 
	int rows = dbi_result_get_numrows(result);
	if( rows == 0) { // new triplet
		dbi_result_free(result);
		syslog(LOG_INFO,"new: %s -> %s, %s",sender(),recipient(),client());
		//since we dont name the fields anymore this has to be in sync with the dbdefs.cpp definition
		stringstream sqlInsert;
		sqlInsert << "INSERT INTO "TABLE_TRIPLETS" VALUES (" << recipient() << ',' << sender() << ',';
		sqlInsert << triplet.getClientAddressNumericStr(',') << ",0," << static_cast<unsigned long>(ltime) << ")";
		if(verbose)
			syslog(LOG_DEBUG,sqlInsert.str().c_str());
		result = DBI_CON_QUERYF( _con, sqlInsert.str().c_str() );
		if(!result)
			throw runtime_error(getConError("update insert: ")+sqlInsert.str() );
		recStatus = unknown;
	} else { // known triplet
		dbi_result_next_row(result); 
		unsigned long count = dbi_result_get_ulong_idx(result,1);
		unsigned long recTimeStamp = dbi_result_get_ulong_idx(result,2);
		if( (count > 0) || ((ltime - recTimeStamp) > g_getCfg().getTimeout()) ) {
			syslog(LOG_INFO,"ok: %s -> %s, %s (%lu, %lu secs)",sender(),recipient(),client(),
				count,(ltime - recTimeStamp));
			dbi_result_free(result);
			recTimeStamp = ltime;
			stringstream sqlUpdate;
			sqlUpdate << "UPDATE "TABLE_TRIPLETS" SET count=count+1,uts=" << static_cast<unsigned long>(ltime);
			sqlUpdate << " WHERE ( " << sqlWhere.str() << " )";
			if(verbose)
				syslog(LOG_DEBUG,sqlUpdate.str().c_str());
			result = DBI_CON_QUERYF(_con,sqlUpdate.str().c_str());
			if(!result)
				throw runtime_error(getConError("update update: ")+sqlUpdate.str() );
			recStatus = registered;
		} else { // waiting triplet
			syslog(LOG_INFO,"wait: %s -> %s, %s (%lu, %lu secs)",sender(),recipient(),client(),
				count,(ltime - recTimeStamp));
			recStatus = waiting;
		}
	}
	if(result)
		dbi_result_free(result);
	return recStatus; 
}
#else /// OLDWEAK
DB::RecordStatus DB::update(const Triplet& triplet) throw(std::exception)
{
// unlikely to happen
//	if(!_con)
//		throw runtime_error(getConError("update"));
	dbi_driver dbidrv = dbi_conn_get_driver(_con);
	DBIQuote sender(dbidrv, triplet.getSender().substr(0,MAX_LEN_SENDER));
	DBIQuote recipient(dbidrv, triplet.getRecipient().substr(0,MAX_LEN_RECIPIENT));
	DBIQuote client(dbidrv, triplet.getClientAddress().substr(0,MAX_LEN_CLIENT));
	string clientNumeric = triplet.getClientAddressNumeric();
//	bool wl_network = g_getCfg().isWlDb(TABLE_NETWORK);
	bool weak = g_getCfg().isModeWeak();
	bool verbose = g_getCfg().isVerbose();
	RecordStatus recStatus = unknown;
	time_t ltime;
	dbi_result result;
	
	if(clientNumeric.size() != IPADDR_NUM)
		throw runtime_error("Invalid IP Address: "+triplet.getClientAddress()
			+ " parsed: "+triplet.getClientAddressNumericStr());
//	DBIQuote weakclient(dbidrv,((weak||wl_network)?trimLastByte(triplet.getClientAddress().substr(0,MAX_LEN_CLIENT)):""));
	DBIQuote weakclient(dbidrv,trimLastByte(triplet.getClientAddress().substr(0,MAX_LEN_CLIENT)));
	// check triplet against wl modules
	if(!_wlMods.empty()) {
		Triplet quotedTriplet(client(),sender(),recipient(),weakclient(),clientNumeric);
		for(WlModuleList::const_iterator itr=_wlMods.begin(); itr != _wlMods.end();++itr) 
			if( (*itr)->check(quotedTriplet) )
				return registered;
	}
	//construct the sql select string
	string sqlSelect("SELECT count,uts FROM "TABLE_TRIPLETS" WHERE ");
	string sqlWhere(string(ATTR_RECIPIENT"=")+recipient()+" AND "ATTR_SENDER"="+sender());
	string sqlOrder;
	if(weak) { // weak greylisting
		sqlWhere += string(" AND "ATTR_CLIENT" RLIKE ") + weakclient();
		sqlOrder = " ORDER BY count DESC";
	} else { // normal greylisting
		sqlWhere += std::string(" AND "ATTR_CLIENT"=") + client();
	}
	sqlSelect += sqlWhere + sqlOrder + " LIMIT 1";
	if(verbose)
			syslog(LOG_DEBUG,sqlSelect.c_str());

	result = DBI_CON_QUERYF(_con, sqlSelect.c_str());
	if(!result)
		throw runtime_error(getConError("update select") );

	time(&ltime); //get ultime (unix timestamp) in ltime 
	int rows = dbi_result_get_numrows(result);
	if( rows == 0) {
		// new triplet
		dbi_result_free(result);
		syslog(LOG_INFO,"new: %s -> %s, %s",sender(),recipient(),client());
		//since we dont name the fields anymore this has to be in sync with the dbdefs.cpp definition
		stringstream stm;
		stm << "INSERT INTO "TABLE_TRIPLETS" VALUES (" << client() << ',' << recipient() << ',' << sender() << ',';
		stm << "0," << static_cast<unsigned long>(ltime) << ")";
		if(verbose)
			syslog(LOG_DEBUG,stm.str().c_str());
		
		result = DBI_CON_QUERYF( _con, stm.str().c_str() );
		if(!result)
			throw runtime_error(getConError("update insert") );
		recStatus = unknown;
	} else {
		// known triplet
		dbi_result_next_row(result); 
		unsigned long count = dbi_result_get_ulong_idx(result,1);
		unsigned long recTimeStamp = dbi_result_get_ulong_idx(result,2);

		if( (count > 0) || ((ltime - recTimeStamp) > g_getCfg().getTimeout()) ) {
			syslog(LOG_INFO,"ok: %s -> %s, %s (%lu, %lu secs)",sender(),recipient(),client(),count,(ltime - recTimeStamp));
			dbi_result_free(result);
			count++;
			if(count == COUNT_RESET) 
				count = 1;
			recTimeStamp = ltime;
			stringstream stm;
			stm << "UPDATE "TABLE_TRIPLETS" SET count=count+1,uts=" << static_cast<unsigned long>(ltime);
			/// TODO this line should be removed after the database has been updated
			///      also note http://dev.mysql.com/doc/mysql/en/REPLACE.html could be faster
			//stm << ',' << triplet.getClientAddressNumericStrSQL(',',"ip",64,"=",2);
			stm << " WHERE " << sqlWhere;
			if(verbose)
				syslog(LOG_DEBUG,stm.str().c_str(),ltime);
			result = DBI_CON_QUERYF(_con,stm.str().c_str(),ltime);
			if(!result)
				throw runtime_error(getConError("update update") );
			recStatus = registered;
		} else { //count is likely to be one => check timediff
			syslog(LOG_INFO,"wait: %s -> %s, %s (%lu, %lu secs)",sender(),recipient(),client(),count,(ltime - recTimeStamp));
			recStatus = waiting;
		}
	}
	dbi_result_free(result);
	return recStatus; 
}
#endif
dbi_result DB::executeSQL(const std::string& sql) throw(std::exception)
{
	if(!_con)
		throw runtime_error(getConError("executeSQL"));
	dbi_result result = DBI_CON_QUERYF(_con, sql.c_str() );
	if(!result) {
		syslog(LOG_INFO,"SQL error: %s",getConError("executeSQL").c_str());
//		throw runtime_error( getConError("executeSQL") );
	}
	return result;
}
void DB::createTable(const WLStaticDef& staticDef) throw(std::exception)
{	
	if(!_con)
		throw runtime_error(getConError("createTable: no db connection"));
	bool verbose = g_getCfg().isVerbose();
	stringstream sqlTestTable;
	sqlTestTable << staticDef._testTable << " WHERE 1=0"; //more portable than " LIMIT 1";
	if(verbose)
		syslog(LOG_DEBUG,sqlTestTable.str().c_str());
	dbi_result result = DBI_CON_QUERYF(_con, sqlTestTable.str().c_str() );
	if(!result) {
		syslog(LOG_INFO,"testing table failed: %s : %s, creating table",
			staticDef._tableName,getConError("").c_str());
		unsigned iCT = 0;
		while( staticDef._createTable[iCT] ) {
			stringstream sqlCreateTable;
			sqlCreateTable << staticDef._createTable[iCT];
			if(staticDef._primaryKey)
				sqlCreateTable << " PRIMARY KEY (" << staticDef._primaryKey << " )";
			sqlCreateTable << ")";
			if(verbose)
				syslog(LOG_DEBUG,sqlCreateTable.str().c_str());
			result = DBI_CON_QUERYF(_con, sqlCreateTable.str().c_str());
			if(!result && (g_getCfg().getDbType() == "sqlite") ) {
				//table created, sqlite driver currently returns empty result nevertheless
				result = DBI_CON_QUERYF(_con, sqlTestTable.str().c_str() );
			}
			if(!result) {
					throw runtime_error( getConError("create table: ")+staticDef._tableName
						+ string( verbose?" SQL: "+sqlCreateTable.str() : "") );
			} else {
				dbi_result_free(result);
				result = 0;
			}
			iCT++;
		}
		syslog(LOG_INFO,"table %s created", staticDef._tableName );
	}
	if(result)
		dbi_result_free(result);
}
std::string DB::getSqlCondition(const WLCompares& compares,const Triplet& triplet) const
{
	string sql;
	int j = 0;
	while(j < MAX_WL_COMPARES) {
		if(compares[j]._compareWith) {
			if(compares[j]._compareBinder)
				sql += std::string(" ") + compares[j]._compareBinder + " ";
			sql += compares[j]._compareWith;
			sql += compares[j]._compareOp;
			sql += triplet.getMember(compares[j]._tripletMember);
			j++;
		} else
			break;
	}
	return sql;
}
bool DB::exactMatch(const std::string& logprefix,const WLStaticDef& staticDef,
	const Triplet& triplet) throw(std::exception)
{
	bool verbose = g_getCfg().isVerbose();
	string sqlWlSelect = staticDef._searchTable;
	sqlWlSelect += " WHERE " + getSqlCondition(staticDef._compares,triplet);
	sqlWlSelect += " LIMIT 1";
	if(verbose)
		syslog(LOG_DEBUG,sqlWlSelect.c_str());
	dbi_result result = DBI_CON_QUERYF(_con, sqlWlSelect.c_str());
	if(!result)
		throw runtime_error(getConError("exactMatch")+staticDef._tableName );
	int rows = dbi_result_get_numrows(result);
	if( rows == 0) {
		dbi_result_free(result);
		return false;
	} else {
		dbi_result_next_row(result); 
		const char *comment = dbi_result_get_string_idx(result,1);
		syslog(LOG_INFO,"%s %s: %s -> %s, %s: %s",
			logprefix.c_str(),staticDef._tableName,triplet.getSender().c_str(),
			triplet.getRecipient().c_str(),triplet.getClientAddress().c_str(),comment);
		dbi_result_free(result);
		return true;
	}
}
const std::string DB::numToStr(unsigned u) const
{
	stringstream stm;
	stm << u;
	return stm.str();
}



